/********************************************************************************
*                                                                               *
* Copyright (c) 2007, Computer Networks and Internet, University of Tübingen,   *
* Sand 13, 72076 Tübingen, Germany                                              *
*                                                                               *
* All rights reserved.                                                          *
*                                                                               *
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU              *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA       *
*                                                                               *
********************************************************************************/
#include "PAMConfiguration.h"

// Standard C includes
#include <stdlib.h>
#include <sys/stat.h>
#include <syslog.h>
#include <cstring>
#include <stdio.h>

// Standard C++ includes
#include <stdexcept>


namespace pam_xacml {


PAMConfiguration::PAMConfiguration() {
	//Defaut Values

	//optional parameters
	debug = 0;
	dumpXacml = false;
	mandatoryObligations = true;
	obligationsIgnore = false;
	applicationName = NULL;
	pdpFormat = NULL;
	ldapConf = NULL;
	ou = NULL;

	use_ssl = false;
	file_ca_chain = NULL;
	file_client_cert_key = NULL;
	file_client_password = NULL;

	//mandatory parameters
	pdpEndpoint = NULL;
	staticResponse = NULL;
	requestResource = NULL;
	requestAction = NULL;

	requestTemplate = NULL;
	obligationsScript = NULL;
	requestBuilderScript = NULL;
}

PAMConfiguration::~PAMConfiguration() {
	/*if (applicationName)
		free(applicationName);
	 */
	if (file_ca_chain)
		free(file_ca_chain);
	if (file_client_cert_key)
		free(file_client_cert_key);
	if (file_client_password)
		free(file_client_password);

	if (pdpEndpoint)
		free(pdpEndpoint);
	if (staticResponse)
		free(staticResponse);
	if (requestResource)
		free(requestResource);
	if (requestAction)
		free(requestAction);

	if (requestTemplate)
		free(requestTemplate);

	if (obligationsScript)
		free(obligationsScript);

	if (requestBuilderScript)
		free(requestBuilderScript);

	if (pdpFormat)
			free(pdpFormat);
	if (ldapConf)
			free(ldapConf);
	if (ou)
			free(ou);
}

char* PAMConfiguration::getValueOfOption(char** savePtr, const char* tok) {
	char* result = strtok_r(NULL, "", savePtr);
	if (result==NULL) {
		throw std::invalid_argument("Value of "+std::string(tok)+" is undefined");
	}
	return strdup(result);
}


/**
 * @returns 0 on error and 1 on success
 */
void PAMConfiguration::parsePamArguments(int argc, const char** argv)
{
    char* tok;
    char* saveptr;

	int i;

	//tries to parse all arguments
	for (i=0; i < argc; i++) {
		//check for presence of single tags
		if (0==strcasecmp(argv[i], "debug")) {
			debug=true;
		} else if (0==strcasecmp(argv[i], "dumpXacml")) {
			dumpXacml=true;


		} else if (0==strcasecmp(argv[i], "obligationsNotMandatory")) {
			mandatoryObligations = false;
		} else if (0==strcasecmp(argv[i], "obligationsIgnore")) {
			obligationsIgnore = true;


		} else {
			//tok is value before '=', next strtok_r is value directly after '='
			tok = strtok_r((char*)argv[i], "=", &saveptr);

			if (0==strcasecmp(tok, "requestBuilder")) {
				char* requestBuilderType = getValueOfOption(&saveptr, tok);
				if (0==strcasecmp(requestBuilderType,"INTERNAL")) {
					requestBuilder = INTERNAL_REQUEST_BUILDER;
				} else if (0==strcasecmp(requestBuilderType,"APPLICATION")) {
					requestBuilder = APPLICATION_REQUEST_BUILDER;
				} else if (0==strcasecmp(requestBuilderType,"REST")) {
					requestBuilder = SOAP_REQUEST_BUILDER;
				} else if (0==strcasecmp(requestBuilderType,"SCRIPT")) {
					requestBuilder = SCRIPT_REQUEST_BUILDER;
				} else {
					throw std::invalid_argument("PAMConfiguration::parsePamArguments unknown requestBuilder: "+std::string(requestBuilderType));
				}
			} else if (0==strcasecmp(tok, "pdpRequester")) {
				char* pdpRequesterType = getValueOfOption(&saveptr, tok);
				if (0==strcasecmp(pdpRequesterType,"FILE")) {
					pdpRequester = FILE_PDP_REQUESTER;
				} else if (0==strcasecmp(pdpRequesterType,"SIMPLE")) {
					pdpRequester = SIMPLE_PDP_REQUESTER;
				} else if (0==strcasecmp(pdpRequesterType,"HANNES")) {
					pdpRequester = HANNES_PDP_REQUESTER;
				} else if (0==strcasecmp(pdpRequesterType, "BESTER")) {
					pdpRequester = BESTER_PDP_REQUESTER;
				} else if (0==strcasecmp(pdpRequesterType, "REST")) {
					pdpRequester = SOAP_PDP_REQUESTER;
				} else {
					throw std::invalid_argument("PAMConfiguration::parsePamArguments unknown pdpRequester: "+std::string(pdpRequesterType));
				}
			}

			/*else if (0==strcasecmp(tok, "appname")) {
				applicationName = getValueOfOption(&saveptr, tok);
		} */else if (0==strcasecmp(tok, "pdpAddress"))
			{
				pdpEndpoint= getValueOfOption(&saveptr, tok);
				if (0==strncasecmp(pdpEndpoint, "https://", 8))
				{
					use_ssl = true;
				}
			} else if (0==strcasecmp(tok, "pdpEndpoint")) {
				pdpEndpoint = getValueOfOption(&saveptr, tok);
			}
			else if (0==strcasecmp(tok, "pdpFormat")) {
				pdpFormat = getValueOfOption(&saveptr, tok);
				if (0!=strcasecmp(pdpFormat, "REST"))
					throw std::invalid_argument("Unknown pdpFormat (REST): " + std::string(argv[i]));
			} else if (0==strcasecmp(tok, "ldapConf")) {
				ldapConf = getValueOfOption(&saveptr, tok);
			} else if (0==strcasecmp(tok, "staticResponse")) {
				staticResponse = getValueOfOption(&saveptr, tok);
			} else if (0==strcasecmp(tok, "requestResource")) {
			    requestResource = getValueOfOption(&saveptr, tok);
			} else if (0==strcasecmp(tok, "requestAction")) {
				requestAction = getValueOfOption(&saveptr, tok);
			}


			//security parameters
			else if (0==strcasecmp(tok, "cachain")) {
				file_ca_chain = getValueOfOption(&saveptr, tok);
			} else if (0==strcasecmp(tok, "clientcertkey")) {
				file_client_cert_key = getValueOfOption(&saveptr, tok);
			} else if (0==strcasecmp(tok, "passphrase")) {
				file_client_password = getValueOfOption(&saveptr, tok);
			}

			//templates
			else if (0==strcasecmp(tok, "requestTemplate")) {
				requestTemplate = getValueOfOption(&saveptr, tok);
			}

			//external scripts
			else if (0==strcasecmp(tok, "obligationsScript")) {
				obligationsScript = getValueOfOption(&saveptr, tok);
			}
			else if (0==strcasecmp(tok, "requestBuilderScript")) {
				requestBuilderScript = getValueOfOption(&saveptr, tok);
			}

			else {
				throw std::invalid_argument("Unknown option in PAM configuration: " + std::string(argv[i]));
			}
		}
	}


	if (use_ssl) {
		readSSLPassphrase(file_client_password, debug, &client_password);
	}


	validateConfiguration();
}

void PAMConfiguration::validateConfiguration() {

	if (obligationsIgnore && obligationsScript) {
		throw std::invalid_argument("You cannot ignore obligations when specifying an obligationsScript.");
	}
	if ((mandatoryObligations && obligationsIgnore)) {
		throw std::invalid_argument("You MUST specify obligationsNotMandatory when using obligationsIgnore");
	}
	if ((pdpRequester!=FILE_PDP_REQUESTER) && (pdpEndpoint == NULL)) {
		throw std::invalid_argument("PAMConfiguration::validateConfiguration no pdpEndpoint");
	}
	if ((pdpRequester==FILE_PDP_REQUESTER) && (staticResponse == NULL)) {
		throw std::invalid_argument("PAMConfiguration::validateConfiguration no staticResponse");
	}
	if ((requestBuilder==SCRIPT_REQUEST_BUILDER) && ! requestBuilderScript) {
		throw std::invalid_argument("If you say SCRIPT for requestBuilder, you have to provide a requestBuilderScript");
	}
	if ((requestTemplate && requestBuilderScript)) {
		throw std::invalid_argument("You may not provide a requestTemplate and a requestBuilderScript. Use only one of the options.");
	}

}

void PAMConfiguration::readSSLPassphrase(char* file_client_password, int debug, char** client_password)
{
	char* buf;
	struct stat size;

	FILE* fh = fopen(file_client_password, "r");
	if (!fh) {
		throw std::invalid_argument("PAMConfiguration::readSSLPassphrase could not read client key passphrase from file " + std::string(file_client_password));
	}

#ifndef __APPLE__
	fstat(fh->_fileno, &size);
#else
	stat(file_client_password, &size);
#endif

	buf = (char*)malloc(size.st_size + 2);
	if (buf != NULL) {
		fread(buf, 1, size.st_size, fh);
		fclose(fh);
		buf[size.st_size-1] = '\0';
		*client_password = strdup(buf);
		if (debug) {
			syslog(LOG_INFO, "  readSSLPassphrase: read passphrase of length %d", strlen(buf));
		}
	} else {
		fclose(fh);
		throw std::bad_alloc();
	}
}



// Getters

const char* PAMConfiguration::getApplicationName()
{
	return this->applicationName;
}

RequestBuilderType PAMConfiguration::getRequestBuilder(){
	return this->requestBuilder;
}
PdpRequesterType PAMConfiguration::getPdpRequester() {
	return this->pdpRequester;
}
const char* PAMConfiguration::getStaticResponse()
{
	return this->staticResponse;
}
const char* PAMConfiguration::getPdpEndpoint(){
	return this->pdpEndpoint;
}

const char* PAMConfiguration::getRequestResource(){
	return this->requestResource;
}
const char* PAMConfiguration::getUsername(){
	return this->username;
}

const char* PAMConfiguration::getRequestAction() {
	return this->requestAction;
}

bool PAMConfiguration::isMandatoryObligations(){
	return this->mandatoryObligations;
}

bool PAMConfiguration::isObligationsIgnore(){
	return this->obligationsIgnore;
}

bool PAMConfiguration::isDebug(){
	return this->debug;
}

bool PAMConfiguration::isUseSSL(){
	return this->use_ssl;
}

const char* PAMConfiguration::getFileCAchain(){
	return this->file_ca_chain;
}
const char* PAMConfiguration::getFileClientCertKey(){
	return this->file_client_cert_key;
}
const char* PAMConfiguration::getFileClientPassword(){
	return this->file_client_password;
}
const char* PAMConfiguration::getClientPassword(){
	return this->client_password;
}

const char* PAMConfiguration::getRequestTemplate(){
	return this->requestTemplate;
}

const char* PAMConfiguration::getObligationsScript() {
	return this->obligationsScript;
}

const char* PAMConfiguration::getRequestBuilderScript() {
	return this->requestBuilderScript;
}

const char* PAMConfiguration::getPDPFormat() {
	return this->pdpFormat;
}

const char* PAMConfiguration::getLdapConf() {
	return this->ldapConf;
}

// Setter

void PAMConfiguration::setUsername(char* user) {
	this->username = user;
}

void PAMConfiguration::setAppname(char* app) {
	this->applicationName = app;
}


}
