/********************************************************************************
*                                                                               *
* Copyright (c) 2007, Computer Networks and Internet, University of Tübingen,   *
* Sand 13, 72076 Tübingen, Germany                                              *
*                                                                               *
* All rights reserved.                                                          *
*                                                                               *
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU              *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA       *
*                                                                               *
********************************************************************************/
#include "libtiming.h"

#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>


#define LIBTIMING_FILENAME_SIZE 150
#define LIBTIMING_FILE_LINELEN 150

/**
 * @param dstFolder PATH without a trailing slash
 */
libTimingSession* libtiming_init(char* appName, char* dstFolder)
{
	libTimingSession* sess = (libTimingSession*)calloc(1, sizeof(libTimingSession));
	sess->appName = appName;
	sess->dstFolder = dstFolder;
	
	time_t now;
	struct tm *locnow;
	time(&now);
	locnow = localtime(&now);
	
	char curTimeString[80];
	snprintf(curTimeString, 80, "%d%02d%02d-%02d%02d%02d", locnow->tm_year+1900, locnow->tm_mon+1, locnow->tm_mday,
				locnow->tm_hour, locnow->tm_min, locnow->tm_sec);

	int file; // File descriptor
	char* filename = (char*)malloc(LIBTIMING_FILENAME_SIZE+1);
	snprintf(filename, LIBTIMING_FILENAME_SIZE, "%s/%s-XXXXXX", dstFolder, curTimeString);
	file = mkstemp(filename);
	free(filename);
	
	char writebuf[LIBTIMING_FILE_LINELEN+1];
	
	snprintf(writebuf, LIBTIMING_FILE_LINELEN, "PROFILING INFORMATION FOR *%s*\n", appName);
	write(file, writebuf, strlen(writebuf));
	snprintf(writebuf, LIBTIMING_FILE_LINELEN, "RUN ON %s\n", curTimeString);
	write(file, writebuf, strlen(writebuf));
	snprintf(writebuf, LIBTIMING_FILE_LINELEN, "FORMAT: delta-time;absolute-time;identifier - ALL TIMES IN MILLISECONDS\n", curTimeString);
	write(file, writebuf, strlen(writebuf));

	sess->filehandle = file;
	
	return sess;
}

void startMeasurement(libTimingSession* sess)
{
	gettimeofday(&(sess->startTime),(struct timezone *)0);
}

void measure(libTimingSession* sess, char* identifier)
{	
	measurePoint* p = (measurePoint*)calloc(1, sizeof(measurePoint));
	gettimeofday(&(p->point),(struct timezone *)0);
	p->identifier = strdup(identifier);
	
	if (sess->last == NULL) {
		sess->first = p;
		sess->last = p;
	} else {
		sess->last->next = p;
		sess->last = p;
	}
}

void endMeasurement(libTimingSession* sess)
{
	gettimeofday(&(sess->endTime),(struct timezone *)0);
}

double getTimeVal(struct timeval val)
{
	return ((double)(val.tv_sec*1000))+(((double)val.tv_usec)/1000);
}

void libtiming_finish(libTimingSession* sess)
{
	char writebuf[LIBTIMING_FILE_LINELEN+1];
	int file = sess->filehandle;
	
	if ((sess->endTime.tv_sec == 0) && (sess->endTime.tv_usec == 0)) {
		// Warning: has not been ended correctly
		endMeasurement(sess);

		snprintf(writebuf, LIBTIMING_FILE_LINELEN, "LIBTIMING-WARNING! PROFILING ENDED WITHOUT A CALL TO endMeasurement()!\n");
		write(file, writebuf, strlen(writebuf));
	}
	
	time_t now;
	struct tm *locnow;
	time(&now);
	locnow = localtime(&now);
		
	char curTimeString[80];
	snprintf(curTimeString, 80, "%d%02d%02d-%02d%02d%02d", locnow->tm_year+1900, locnow->tm_mon+1, locnow->tm_mday,
			locnow->tm_hour, locnow->tm_min, locnow->tm_sec);
		
	snprintf(writebuf, LIBTIMING_FILE_LINELEN, "PROFILING ENDED ON %s\n\n", curTimeString);
	write(file, writebuf, strlen(writebuf));
	
	measurePoint* walker = sess->first;
	double lastPoint = getTimeVal(sess->startTime);
	double currentPoint = getTimeVal(sess->startTime);
	
	// START ENTRY
	snprintf(writebuf, LIBTIMING_FILE_LINELEN, "%.3f;%.3f;START OF MEASUREMENT\n", currentPoint-lastPoint, currentPoint);
	write(file, writebuf, strlen(writebuf));
	
	while (walker) {
		// FILE FORMAT:
		// deltatime;real_time;Identifier
		currentPoint = getTimeVal(walker->point);
		snprintf(writebuf, LIBTIMING_FILE_LINELEN, "%.3f;%.3f;%s\n", currentPoint-lastPoint, currentPoint, walker->identifier);
		write(file, writebuf, strlen(writebuf));
		lastPoint = currentPoint;
		walker = walker->next;
	}
	
	currentPoint = getTimeVal(sess->endTime);
	// EXIT ENTRY
	snprintf(writebuf, LIBTIMING_FILE_LINELEN, "%.3f;%.3f;END OF MEASUREMENT\n", currentPoint-lastPoint, currentPoint);
	write(file, writebuf, strlen(writebuf));	
	
	close(sess->filehandle);

	
	// Free measure-points
	measurePoint* p = sess->first;
	measurePoint* tmp;
	while (p) {
		tmp = p->next;
		free(p->identifier);
		free(p);
		p=tmp;
	}
	
	free(sess);
}