/********************************************************************************
*                                                                               *
* Copyright (c) 2007, Computer Networks and Internet, University of Tübingen,   *
* Sand 13, 72076 Tübingen, Germany                                              *
*                                                                               *
* All rights reserved.                                                          *
*                                                                               *
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU              *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA       *
*                                                                               *
********************************************************************************/
#define GNU_SOURCE		// Needed for get_current_dir_name

// pam_xacml includes
#include "config.h"
#include "pam_handler.h"
#include "pam_xacml_constants.h"
#include "PAMConfiguration.h"

//do the work on XACML
#include "xacml.h"
//#include "requestBuilders/ApplicationAsRequestBuilder.h"
#include "requestBuilders/PAMRequestBuilder.h"
//#include "requestBuilders/ScriptRequestBuilder.h"
#include "pdpRequesters/pdpRequesterFactory.h"


// Standard C includes
#include <stdlib.h>
#include <sys/stat.h>

#include <syslog.h>

// Standard C++ includes
#include <stdexcept>

// PAM Includes
#ifdef __APPLE__
extern "C" {
	#include <pam/pam_modules.h>
	#include <pam/pam_misc.h>
	#include <pam/_pam_macros.h>
}
#else
	#include <security/pam_modules.h>
	#include <security/pam_misc.h>
	#include <security/_pam_macros.h>
#endif


//#define PROFILING
#ifdef PROFILING
	#include "libtiming.h"
#endif

using namespace pam_xacml;

/**
 * Writes the request to a temporary file and outputs it's name to syslog
 *
 * @returns 0 on failure, 1 on success.
 */
int writeXacmlRequestToTemporaryFile(char* request)
{
	///////////////////////////////////////////////////////////////////////////////
	// Shall the Request be written to a temporary file for debugging purposes? //
	/////////////////////////////////////////////////////////////////////////////
	int xacmlFile; // File descriptor
	char* tmpFile = (char*)malloc(50);
	strcpy(tmpFile, "/tmp/xacmlXXXXXX");
	xacmlFile = mkstemp(tmpFile);
	if (xacmlFile != -1) {
		syslog (LOG_INFO, " - opened temporary file %s for XACML-Request output", tmpFile);
		write(xacmlFile, request, strlen(request));
		close(xacmlFile);
		free(tmpFile);
		return 1;
	} else {
		free(tmpFile);
		syslog (LOG_ERR, " could not output XACML-Request to temporary directory.");
		return 0;
	}
}





extern "C" int pam_authorize(pam_handle_t *pamh, int flags, int argc, const char **argv) {
	return authorize(pamh, flags, argc, argv);
}

bool passObligationsToScript(PAMConfiguration* conf, char* obligations)
{
	FILE* app;
	int status;
	std::string returnValue;
	char* bufPtr = obligations;
	app = popen((std::string(conf->getObligationsScript())).c_str(), "w");
	if (app == NULL)
		throw std::runtime_error("Could not invoke application that was specified in obligationsScript");

	while(*bufPtr) {
		bufPtr += fwrite(bufPtr, 1, strlen(bufPtr), app);
	}

	status = pclose(app);

	syslog(LOG_INFO, "Exit status of obligationsScript was %d", WEXITSTATUS(status));

	if (WEXITSTATUS(status) == 0) {
		return true;
	} else {
		return false;
	}
}

int authorize(pam_handle_t *pamh, int flags, int argc, const char **argv)
{
	int resultCode;
	struct pam_conv* conv;
	char* xacmlRequest = NULL;
	char* xacmlReply;

	// Fields needed for XACML request
	char* username;
	char* appname;

	//for each request its own context
#ifdef PROFILING
	libTimingSession* sess;
	sess = libtiming_init("PAMXACML", "/tmp/profiling-pamxacml");
	startMeasurement(sess);
#endif

	openlog(PACKAGE, 0, LOG_AUTHPRIV);

#ifdef PROFILING
	measure(sess, "Initialization of structures and logging complete");
#endif


	PAMConfiguration myConf;
#ifndef DIRTY_CONF_HACK
	try {
		myConf.parsePamArguments(argc, argv);
	} catch (std::exception &exc) {
		syslog(LOG_ERR, "error parsing the parameters provided by the pam-module from /etc/pam.d/APP. Reason: %s",exc.what());
		closelog();
		return PAM_AUTH_ERR;
	}
#else

#endif


#ifdef PROFILING
	measure(sess, "Read arguments from /etc/pam.d-Configuration file");
#endif

	/////////////////////////////////////////////////////////////////////////
	/// First we get some information from the PAM Library
	/////////////////////////////////////////////////////////////////////////

	// Get user name
	resultCode = pam_get_item(pamh, PAM_USER, (const void**)&username);
	if (resultCode != PAM_SUCCESS) {
		syslog(LOG_ERR, "Could not get username.");
		closelog();
		return PAM_USER_UNKNOWN;
	}
	else
		myConf.setUsername(username);

	// Get application name
	resultCode = pam_get_item(pamh, PAM_SERVICE, (const void**)&appname);
	if (resultCode != PAM_SUCCESS) {
		syslog(LOG_ERR, "Could not get service name.");
		closelog();
		return PAM_USER_UNKNOWN;
	}
	else
		myConf.setAppname(appname);


	// Get conversation function
	resultCode = pam_get_item(pamh, PAM_CONV, (const void **)&conv);
	if (resultCode != PAM_SUCCESS) {
		syslog(LOG_ERR, "Could not get conversation function.");
		closelog();
		return PAM_AUTH_ERR;
	}

	if (myConf.isDebug()) {
		syslog(LOG_INFO, " - This is %s, Version %s - bug reports to %s", PACKAGE, PACKAGE_VERSION, PACKAGE_BUGREPORT);
		syslog(LOG_INFO, " - DEBUG=%d, APPNAME=%s, PDP=%s", myConf.isDebug(), myConf.getApplicationName(), myConf.getPdpEndpoint());
		syslog(LOG_INFO, " - Parsing configuration file.");
	}
	/////////////////////////////////////////////////////////////////////////
	/// Get the XACML Request. Either from Application or build internally
	/////////////////////////////////////////////////////////////////////////
#ifdef PROFILING
	measure(sess, " - Before get XACML Request");
#endif
	syslog(LOG_INFO, " - Building request");

	try {
		/*
		if (myConf.getRequestBuilder() == APPLICATION_REQUEST_BUILDER) {
			ApplicationAsRequestBuilder arp(pamh, conv, myConf.isDebug());
			xacmlRequest = (char*)arp.buildAuthzRequest();
			syslog(LOG_INFO, " - got result from application: %s",xacmlRequest);
		} else if (myConf.getRequestBuilder() == INTERNAL_REQUEST_BUILDER) {
			AttributeProvider ap(username, appname, &guid);
			PAMRequestBuilder prb(&ap, &myConf);
			xacmlRequest = (char*)prb.buildAuthzRequest();
			syslog(LOG_INFO, " - built XACML request from template: %s",xacmlRequest);
		} else if (myConf.getRequestBuilder() == SOAP_REQUEST_BUILDER) {
			AttributeProvider ap(username, appname, &guid);
			PAMRequestBuilder prb(&ap, &myConf);
			syslog(LOG_INFO, " - building XACML request with libzxid");
		} else if (myConf.getRequestBuilder() == SCRIPT_REQUEST_BUILDER) {
			ScriptRequestBuilder srb(&myConf, &guid);
			xacmlRequest = strdup((char*)srb.buildAuthzRequest());
			syslog(LOG_INFO, " - built XACML request with script: %s",xacmlRequest);
		} else {
			syslog(LOG_ERR, " - unknown request builder - Aborting!");
			closelog();
			return PAM_AUTH_ERR;
		}
		*/
		if (myConf.getRequestBuilder() == SOAP_REQUEST_BUILDER)
		{
			AttributeProvider ap(username, appname);
			PAMRequestBuilder prb(&ap, &myConf);
		}
		else
		{
			syslog(LOG_ERR, " - unknown request builder - Aborting!");
			closelog();
			return PAM_AUTH_ERR;
		}
	}
	catch (std::exception &exc) {
		syslog(LOG_ERR, "   failed to build XACML request. REASON: %s", exc.what());
		closelog();
		return PAM_AUTH_ERR;
	}
#ifdef PROFILING
	measure(sess, "After get XACML Request");
#endif



	/////////////////////////////////////////////////////////////////////////
	/// PREPARE AND MAKE XACML REQUEST
	/////////////////////////////////////////////////////////////////////////
#ifdef PROFILING
	measure(sess, "About to talk to the PDP");
#endif
	try {
		PDPRequesterFactory prf(&myConf);
		PDPRequester* PDPr = prf.getPdpRequester(&myConf);
		xacmlReply = PDPr->getPdpDecision(xacmlRequest);

		free(xacmlRequest);
		free(PDPr);
	} catch (std::exception &exc) {
		syslog(LOG_ERR, " - Error talking to PDP - denying access. REASON: %s", exc.what());
		closelog();
		return PAM_AUTH_ERR;
	}
#ifdef PROFILING
	measure(sess, "Done talking to the PDP");
#endif

#ifdef PROFILING
	measure(sess, "About to parse the reply");
#endif

	int decision = NULL;
	if (strstr(xacmlReply, "NotApplicable"))
		decision = XACML_NOT_APPLICABLE;
	else if (strstr(xacmlReply, "Indeterminate"))
		decision = XACML_INDETERMINATE;
	else if (strstr(xacmlReply, "Deny"))
		decision = XACML_DENY;
	else if (strstr(xacmlReply, "Permit"))
		decision = XACML_PERMIT;
	/**
	 * Be pessimistic like usual
	 */
	//if (xacmlReply)
	if(!decision)
	{
		if (strcmp(xacmlReply,"Deny") == 0)
		{
			syslog(LOG_ERR, " - XACMLReply: %s", xacmlReply);
		}
		else if (!(strstr(xacmlReply, "XACMLSAML")))
		{
			try {
				XMLDocument doc(xacmlReply);// parse the reply
				XACMLReply reply(&doc);
				if (1 || myConf.isDebug()) {
				   // dump the resulting document
					//syslog(LOG_INFO, " string representation of reply: %s", xacmlReply);
				   //xmlDocDump(stdout, doc.getDoc());
				}

				#ifdef PROFILING
					measure(sess, "Parsed the reply");
				#endif


				decision = reply.getDecision();
				if (decision == XACML_PERMIT) {
					syslog(LOG_INFO, "XacmlPermit");

					//Obligations are mandatory for the app to understand
					if (myConf.isMandatoryObligations() && reply.hasObligations()) {
						syslog(LOG_INFO, " - XACML Decision with mandatory Obligations");

						resultCode = converseWithApp(pamh, conv, PAM_AUTHZ_RESPONSE_MANDATORY, xacmlReply, NULL, myConf.isDebug());

						if (!resultCode) {
							if (!myConf.getObligationsScript()) {
								// If a obligations script exists, try to use it before ultimatively denying access.
								syslog(LOG_INFO, " - Mandatory Response: Application did not understand me!");
								decision=PAM_XACML_DENY;
							}
						}
					} else {
						if (!myConf.isObligationsIgnore()) {
							syslog(LOG_INFO, " Informational Response: XACML Decision");
							resultCode = converseWithApp(pamh, conv, PAM_AUTHZ_RESPONSE_INFORMATIONAL, xacmlReply, NULL, myConf.isDebug());
						} else {
							syslog(LOG_INFO, " - WARNING!!! Ignoring obligations from PDP as specified in configuration file.");
						}
					}

					if (myConf.getObligationsScript()) {
						if (!passObligationsToScript(&myConf, xacmlReply)) {
							syslog(LOG_ERR, " - ObligationsScript did not understand the obligations.");
							if (myConf.isMandatoryObligations())
								decision=PAM_XACML_DENY;
						}
					}

				} else {
					xmlNodePtr statusNode = reply.getStatus();
					xmlBufferPtr xmlBuf = xmlBufferCreate();
					xmlNodeDump(xmlBuf, doc.getDoc(), statusNode, 2, 1);
					syslog(LOG_ERR, " - Negative authorization decision! XACML status %s", xmlBuf->content);
					xmlBufferFree(xmlBuf);
				}
			} catch (std::invalid_argument &exc) {
			   syslog(LOG_ERR, " - Parsing the XACML reply failed! Reason: %s",exc.what());//TODO add status
			} catch (std::runtime_error &exc) {
			   syslog(LOG_ERR, " - Parsing the XACML reply failed! Reason: %s",exc.what());//TODO add status
		   } catch (...) {
			   syslog(LOG_ERR, " - Parsing the XACML reply failed!");//TODO add status
		   }

			#ifdef PROFILING
				measure(sess, " - Returned obligations and made authorization decision");
			#endif

				//clean up
				free(xacmlReply);


			#ifdef PROFILING
				endMeasurement(sess);
				libtiming_finish(sess);
			#endif
		}
		/**
		 *
		 * Parsage Manuel
		 *
		 */
		else
		{
/*			char* xacmlResponse=NULL;
			int len = strlen(xacmlReply)-strlen("XACMLSAML");

			xacmlResponse = (char*)calloc(len+1, 1);
			syslog(LOG_INFO, "xacmlReply %s", xacmlReply);
			strncat(xacmlResponse,xacmlReply+strlen("XACMLSAML"),strlen(xacmlReply));
			syslog(LOG_INFO, "xacmlResponse %s", xacmlResponse);

			char* begin = strstr(xacmlResponse, ":Decision>");
			char* end = strstr(xacmlResponse, "</xacml-context:Decision>");
			char* response = NULL;
			syslog(LOG_INFO, "Begin %s", begin);
			response = (char*) calloc((strlen(begin)-strlen(end)+1), 1);
			strncat(response, begin, strlen(begin)-strlen(end));
			//syslog(LOG_INFO, "Response %s", response);

			if (strstr(response, "Permit"))
				decision = XACML_PERMIT;
			//syslog(LOG_INFO, "Decision %d", decision);

			free(response);
			//free(xacmlResponse);
*/
			if (strstr(xacmlReply, "NotApplicable"))
				decision = XACML_NOT_APPLICABLE;
			else if (strstr(xacmlReply, "Indeterminate"))
				decision = XACML_INDETERMINATE;
			else if (strstr(xacmlReply, "Deny"))
				decision = XACML_DENY;
			else if (strstr(xacmlReply, "Permit"))
				decision = XACML_PERMIT;
		}
	}

	if (decision==XACML_PERMIT) {
		syslog(LOG_INFO, "Allowed access for user %s", username);
		closelog();
		return PAM_SUCCESS;
	} else {
		syslog(LOG_INFO, "DENIED access for user %s", username);
		closelog();
		return PAM_AUTH_ERR;
	}
}




/**
 * Generic function for conversation with application
 *
 * @param pamh The PAM-Handle
 * @param conv The conversation-structure from PAM
 * @param messageType The type of message
 * @param messageContent The message itself, usually the authorization interface type
 *          (may be NULL - then a String containing only
 * 			\0 is send as content to the Application)
 *
 * @returns 0 on failure, 1 on success
 */
int converseWithApp(pam_handle_t *pamh, struct pam_conv* conv, const int messageType, const char* messageContent, char** responseMsg, const int debug)
{
#define NUM_MESSAGES 1
	// Call the conversation function
	struct pam_response* resp;
	struct pam_message* messagePointer[1];
	struct pam_message getRessource;
	messagePointer[0] = &getRessource;

	int replies = 0;

	if (messageContent == NULL) {
		messageContent = "\0";
	}

	getRessource.msg_style = messageType;
	getRessource.msg = messageContent;

	// do the actual conversation with the application
	replies = conv->conv(NUM_MESSAGES, (const struct pam_message**)messagePointer, &resp, conv->appdata_ptr);

	// ... and parse the response...
	if (replies == PAM_SUCCESS) {
		if (debug) {
			syslog(LOG_INFO, " - conversation - got reply on my request (typ=%d, msg=%s)", messageType, messageContent);
			syslog(LOG_INFO, " - reply: success=%d, reply=%s", resp[0].resp_retcode, resp[0].resp);
		}

  	    //no response
		if (resp[0].resp == NULL) {
			syslog(LOG_INFO, " - reply: did not receive resp");
			_pam_drop_reply(resp, NUM_MESSAGES);
			return 0;
		} else {
			if (responseMsg)
				*responseMsg =  strdup(resp[0].resp);//return string from app if requested
		}

		//TODO  _pam_drop_reply(resp, NUM_MESSAGES);
		return 1;
	}

	syslog(LOG_ERR, " - conversation - failure to reply to my request on (typ=%d, msg=%s)", messageType, messageContent);
	return 0;
}


