/********************************************************************************
*                                                                               *
* Copyright (c) 2007, Computer Networks and Internet, University of Tübingen,   *
* Sand 13, 72076 Tübingen, Germany                                              *
*                                                                               *
* All rights reserved.                                                          *
*                                                                               *
* This library is free software; you can redistribute it and/or                 *
* modify it under the terms of the GNU Lesser General Public                    *
* License as published by the Free Software Foundation; either                  *
* version 2.1 of the License, or (at your option) any later version.            *
*                                                                               *
* This library is distributed in the hope that it will be useful,               *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU              *
* Lesser General Public License for more details.                               *
*                                                                               *
* You should have received a copy of the GNU Lesser General Public              *
* License along with this library; if not, write to the Free Software           *
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA       *
*                                                                               *
********************************************************************************/
/**
 * @file xacml.c
 */

#include "xacml.h"

// Standard C includes
#include <stdlib.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>

//C++ includes
#include <stdexcept>


//libxml2 includes
#include <libxml/tree.h>
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/xpathInternals.h>


namespace pam_xacml {

/*****************************************************************************
 *** FUNCTIONS FOR XML and DOM                                             ***
 *****************************************************************************/

XMLDocument::XMLDocument(const char* xmlString) {
    xmlInitParser();

    _doc = xmlParseMemory(xmlString, strlen(xmlString));
    if (_doc == NULL) {
    	throw std::invalid_argument("XMLDocument::XMLDocument Constructor Error: unable to parse string"+ std::string(xmlString));
    }
/*    xmlNodePtr root;
   	root=xmlDocGetRootElement(*doc);
    xmlReconciliateNs(*doc, root);*/
}

XMLDocument::~XMLDocument() {
    xmlFreeDoc(_doc);
    //xmlCleanupParser(); //can give headache in multi threaded environments, especially with axis
}



int XMLDocument::executeXpathExpression(const xmlChar* xpathExpr, xmlXPathObjectPtr* xpathObj) {
    xmlXPathContextPtr xpathCtx;

    // Create xpath evaluation context
    xpathCtx = xmlXPathNewContext(_doc);
    if(xpathCtx == NULL) {
    	throw std::runtime_error("XMLDocument::executeXpathExpression Error: unable to create new XPath context\n");
    }

    xmlNodePtr root;
    root=xmlDocGetRootElement(_doc);
    xpathCtx->node = root;

   	// Get all namespaces from root element and put into XPath context
    xpathCtx->namespaces = xmlGetNsList(_doc, root);
    xpathCtx->nsNr = 0;
    if (xpathCtx->namespaces != NULL) {
        while (xpathCtx->namespaces[xpathCtx->nsNr] != NULL) {
 	       //printf("prefix->namespace %s -> %s\n", xpathCtx->namespaces[xpathCtx->nsNr]->prefix, xpathCtx->namespaces[xpathCtx->nsNr]->href );
 	       xpathCtx->nsNr++;
        }
    }

   	/*//XACML namepsaces
   	xmlXPathRegisterNs( xpathCtx, NULL, BAD_CAST "urn:oasis:names:tc:xacml:2.0:policy:schema:os");
    xmlXPathRegisterNs( xpathCtx, BAD_CAST "xsi", BAD_CAST "http://www.w3.org/2001/XMLSchema-instance" );
    xmlXPathRegisterNs( xpathCtx, BAD_CAST "policy", BAD_CAST "urn:oasis:names:tc:xacml:2.0:policy:schema:os" );*/


    // Evaluate xpath expression
    *xpathObj = xmlXPathEval (xpathExpr, xpathCtx);

    if(*xpathObj == NULL) {
    	throw std::invalid_argument("XMLDocument::executeXpathExpression Error: unable to evaluate xpath expression " + std::string((const char*)xpathExpr) );
        xmlXPathFreeContext(xpathCtx);
        return(-1);
    }


    //xmlXPathDebugDumpObject(stdout, *xpathObj, 0);

    // Cleanup
    xmlXPathFreeContext(xpathCtx);

    return(0);
}


// get first matching node
xmlNodePtr XMLDocument::searchNodeByName(const xmlNodePtr node, const char* name)
{
    xmlNodePtr cur_node = NULL;


    for (cur_node = node; cur_node; cur_node = cur_node->next) {
        if (cur_node->type == XML_ELEMENT_NODE) {
            //printf("node type: Element, search [%s] name: [%s]\n", name, cur_node->name);
            if (strstr((char*)cur_node->name, name)) {
            	return cur_node;
            }
        }

    }

    // not found
    throw std::invalid_argument("XMLDocument::searchNodeByName Error: could not determine requested node");
}


// get first matching text node
int XMLDocument::getFirstTextNode(const xmlNodePtr node, char **text)
{
    xmlNodePtr cur_node = NULL;

    for (cur_node = node; cur_node; cur_node = cur_node->next) {
        if (cur_node->type == XML_TEXT_NODE) {
            //printf("node type: TEXT_NODE, text: [%s]\n", cur_node->content);
            *text = strdup((char*)cur_node->content);
          	return 1;
        }

    }
    return 0;
}

/*****************************************************************************
 *** FUNCTIONS FOR XACML-REPLIES                                           ***
 *****************************************************************************/

XACMLReply::XACMLReply(XMLDocument* xmlDocument) {
	_xmlDocument = xmlDocument;
}

bool XACMLReply::hasObligations() {
	xmlNodePtr root=xmlDocGetRootElement(_xmlDocument->getDoc());
	bool result;
	int size;

	// for uncompliant mode
	xmlXPathObjectPtr xpathObj = NULL;
	
	// Get all namespaces from root element and put into XPath context
	xmlNsPtr* ns = xmlGetNsList(_xmlDocument->getDoc(), root);
	if (ns != NULL) {
		// Compliant mode
		xmlXPathContextPtr xpathCtx = xmlXPathNewContext(_xmlDocument->getDoc());
		xpathCtx->node = root;

		for ( int i = 0; ns[i] != NULL; i++) {
		   int rv = xmlXPathRegisterNs(xpathCtx, (const char*)ns[i]->prefix == NULL ? (xmlChar*) "default" : ns[i]->prefix, ns[i]->href);
		}

		xmlXPathObjectPtr obj = xmlXPathEval((xmlChar*)("/default:Response/default:Result/policy:Obligations/policy:Obligation"), xpathCtx);
		xmlNodeSetPtr nodeSet = obj->nodesetval;
	
		size = xmlXPathNodeSetGetLength(nodeSet);
	} else {
		// Uncompliant mode
		_xmlDocument->executeXpathExpression((xmlChar*)"//Obligations", &xpathObj);
		size = xmlXPathNodeSetGetLength(xpathObj->nodesetval);
	}
	if (size>0) {
		syslog(LOG_INFO, "Has %d obligation(s)\n", size);
		result = true;
	} else {
		syslog(LOG_INFO, "Has no obligations\n");
		result = false;
	}
	
	if (xpathObj!=NULL) {
	xmlXPathFreeObject(xpathObj);
	}
	return result;
}

int XACMLReply::getDecision() {
	int decision;

	// The following code will work for conformant XACML responses using
	// an Xpath query. Some lines below, there is a fallback code for use
	// with unconformat replies as they are returned by the SUN XACML Framework
	// (and thus unfortunately also with SimplePDP).

	//xmlDocDump(stdout, _xmlDocument->getDoc());
	xmlNodePtr root=xmlDocGetRootElement(_xmlDocument->getDoc());
	
	// Get all namespaces from root element and put into XPath context
	xmlNsPtr* ns = xmlGetNsList(_xmlDocument->getDoc(), root);
	if (ns != NULL) {
		xmlXPathContextPtr xpathCtx = xmlXPathNewContext(_xmlDocument->getDoc());
		xpathCtx->node = root;

		for ( int i = 0; ns[i] != NULL; i++) {
		   int rv = xmlXPathRegisterNs(xpathCtx, (const char*)ns[i]->prefix == NULL ? (xmlChar*) "default" : ns[i]->prefix, ns[i]->href);
		}
		xmlXPathObjectPtr obj = xmlXPathEval((xmlChar*)("/default:Response/default:Result/default:Decision"), xpathCtx);
		xmlNodeSetPtr nodeSet = obj->nodesetval;
	
		if (xmlXPathNodeSetGetLength(nodeSet) > 0) {
			//printf(__FILE__ " - AT NODE %s\n", nodeSet->nodeTab[0]->children->content);
	
			if (strcmp((char*)nodeSet->nodeTab[0]->children->content,"Permit")==0)
				decision = XACML_PERMIT;
			else if (strcmp((char*)nodeSet->nodeTab[0]->children->content,"Deny")==0)
				decision = XACML_DENY;
			else if (strcmp((char*)nodeSet->nodeTab[0]->children->content,"Indeterminate")==0)
				decision = XACML_INDETERMINATE;
			else //NotApplicable
				decision = XACML_NOT_APPLICABLE ;
	
			return decision;
		}
	}

	//printf(__FILE__ " - fallback to uncompliant mode to parse response\n");
	// Fallback Code for Simple PDP
	xmlNodePtr matchingNode, decisionNode;
	
	try {
		//printf(__FILE__ " - inside try block\n");
		matchingNode = _xmlDocument->searchNodeByName(root, "Response");
		matchingNode = _xmlDocument->searchNodeByName(matchingNode->children, "Result");
		decisionNode = _xmlDocument->searchNodeByName(matchingNode->children, "Decision");

		char *decisionStr;
		_xmlDocument->getFirstTextNode(decisionNode->children, &decisionStr);
		//printf(__FILE__ " - decision string: %s\n", decisionStr);

		if (strcmp(decisionStr,"Permit")==0)
			decision = XACML_PERMIT;
		else if (strcmp(decisionStr,"Deny")==0)
			decision = XACML_DENY;
		else if (strcmp(decisionStr,"Indeterminate")==0)
			decision = XACML_INDETERMINATE;
		else //NotApplicable
			decision = XACML_NOT_APPLICABLE ;
	} catch (...) {
		throw std::invalid_argument("XACMLReply::getDecision ERROR: Could not determine decision in XML document");
	}

	return decision;									
}

xmlNodePtr XACMLReply::getStatus() {
	//libxml2 XPATHs either work in documents without any namespace prefix or prefix for all elements
	//our document is broke, so we have to get the decision by hand
	//		executeXpathExpression(doc, "//xmlns:Result/xmlns:Decision", xpathObj);
	//		execute_xpath_expression(doc, "Result/Decision");

    xmlNodePtr root, matchingNode, decisionNode;
   	root = xmlDocGetRootElement(_xmlDocument->getDoc());

   	try {
	   	matchingNode = _xmlDocument->searchNodeByName(root, "Response");
	   	matchingNode = _xmlDocument->searchNodeByName(matchingNode->children, "Result" );
	   	return _xmlDocument->searchNodeByName(matchingNode->children, "Status");
	} catch (...) {
		throw std::invalid_argument("XACMLReply::getStatus ERROR: Could not determine status in XML document");
	}

}

}
